# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from decimal import Decimal


class Migration(migrations.Migration):

    dependencies = [
        ('cc', '0004_txid_not_uniq'),
    ]

    operations = [
        migrations.AddField(
            model_name='currency',
            name='dust',
            field=models.DecimalField(default=Decimal('0.0000543'), verbose_name='Dust', max_digits=18, decimal_places=8),
            preserve_default=True,
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.6 on 2016-05-19 16:52
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('cc', '0005_currency_dust'),
    ]

    operations = [
        migrations.AlterField(
            model_name='currency',
            name='api_url',
            field=models.CharField(blank=True, default='http://localhost:8332', max_length=100, null=True, verbose_name='API hostname'),
        ),
        migrations.AlterField(
            model_name='currency',
            name='label',
            field=models.CharField(default='Bitcoin', max_length=20, unique=True, verbose_name='Label'),
        ),
        migrations.AlterField(
            model_name='currency',
            name='magicbyte',
            field=models.CommaSeparatedIntegerField(default='0,5', max_length=10, verbose_name='Magicbytes'),
        ),
        migrations.AlterField(
            model_name='currency',
            name='ticker',
            field=models.CharField(default='BTC', max_length=4, primary_key=True, serialize=False, verbose_name='Ticker'),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('cc', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='wallet',
            name='label',
            field=models.CharField(max_length=100, unique=True, null=True, verbose_name='Label', blank=True),
        ),
    ]



from django.conf import settings
from decimal import Decimal

CC_CONFIRMATIONS = getattr(settings, 'CC_CONFIRMATIONS', 2)
CC_ADDRESS_QUEUE = getattr(settings, 'CC_ADDRESS_QUEUE', 20)
CC_ALLOW_NEGATIVE_BALANCE = getattr(settings, 'CC_ALLOW_NEGATIVE_BALANCE', Decimal('0.001'))
CC_ACCOUNT = getattr(settings, 'CC_ACCOUNT', '')
CC_ALLOWED_HOSTS = getattr(settings, 'CC_ALLOWED_HOSTS', ['localhost', '127.0.0.1'])



from pycoin.encoding import a2b_hashed_base58, EncodingError


def validate(address, magic_bytes):
	try:
		return a2b_hashed_base58(address)[:1] in bytes(map(int, magic_bytes.split(',')))
	except EncodingError:
		return False



from __future__ import absolute_import
from socket import error as socket_error
from decimal import Decimal
from collections import defaultdict
from http.client import CannotSendRequest

from celery import shared_task
from celery.utils.log import get_task_logger
from bitcoinrpc.authproxy import AuthServiceProxy

from django.db import transaction

from .models import (Wallet, Currency, Transaction, Address,
                       WithdrawTransaction, Operation)
from . import settings
from .signals import post_deposite

logger = get_task_logger(__name__)


@shared_task(throws=(socket_error,))
@transaction.atomic
def query_transactions(ticker=None):
    if not ticker:
        for c in Currency.objects.all():
            query_transactions.delay(c.ticker)
        return

    currency = Currency.objects.select_for_update().get(ticker=ticker)
    coin = AuthServiceProxy(currency.api_url)
    current_block = coin.getblockcount()
    processed_transactions = []

    block_hash = coin.getblockhash(currency.last_block)
    transactions = coin.listsinceblock(block_hash)['transactions']

    for tx in transactions:
        if tx['txid'] in processed_transactions:
            continue

        if tx['category'] not in ('receive', 'generate', 'immature'):
            continue

        process_deposite_transaction(tx, ticker)
        processed_transactions.append(tx['txid'])

    currency.last_block = current_block
    currency.save()

    for tx in Transaction.objects.filter(processed=False, currency=currency):
        query_transaction(ticker, tx.txid)


@transaction.atomic
def process_deposite_transaction(txdict, ticker):
    if txdict['category'] not in ('receive', 'generate', 'immature'):
        return

    try:
        address = Address.objects.select_for_update().get(address=txdict['address'])
    except Address.DoesNotExist:
        return

    currency = Currency.objects.get(ticker=ticker)

    try:
        wallet = Wallet.objects.select_for_update().get(addresses=address)
    except Wallet.DoesNotExist:
        wallet, created = Wallet.objects.select_for_update().get_or_create(
            currency=currency,
            label='_unknown_wallet'
        )
        address.wallet = wallet
        address.save()

    tx, created = Transaction.objects.select_for_update().get_or_create(txid=txdict['txid'], address=txdict['address'], currency=currency)

    if tx.processed:
        return

    if created:
        if txdict['confirmations'] >= settings.CC_CONFIRMATIONS and txdict['category'] != 'immature':
            Operation.objects.create(
                wallet=wallet,
                balance=txdict['amount'],
                description='Deposite',
                reason=tx
            )
            wallet.balance += txdict['amount']
            wallet.save()
            tx.processed = True
        else:
            Operation.objects.create(
                wallet=wallet,
                unconfirmed=txdict['amount'],
                description='Unconfirmed',
                reason=tx
            )
            wallet.unconfirmed += txdict['amount']
            wallet.save()

    else:
        if txdict['confirmations'] >= settings.CC_CONFIRMATIONS and txdict['category'] != 'immature':
            Operation.objects.create(
                wallet=wallet,
                unconfirmed=-txdict['amount'],
                balance=txdict['amount'],
                description='Confirmed',
                reason=tx
            )
            wallet.unconfirmed -= txdict['amount']
            wallet.balance += txdict['amount']
            wallet.save()
            tx.processed = True

    post_deposite.send(sender=process_deposite_transaction, instance=wallet)
    tx.save()


@shared_task(throws=(socket_error,))
@transaction.atomic
def query_transaction(ticker, txid):
    currency = Currency.objects.select_for_update().get(ticker=ticker)
    coin = AuthServiceProxy(currency.api_url)
    for txdict in normalise_txifno(coin.gettransaction(txid)):
        process_deposite_transaction(txdict, ticker)


def normalise_txifno(data):
    arr = []
    for t in data['details']:
        t['confirmations'] = data['confirmations']
        t['txid'] = data['txid']
        t['timereceived'] = data['timereceived']
        t['time'] = data['time']
        arr.append(t)
    return arr


@shared_task()
def refill_addresses_queue():
    for currency in Currency.objects.all():
        coin = AuthServiceProxy(currency.api_url)
        count = Address.objects.filter(currency=currency, active=True, wallet=None).count()

        if count < settings.CC_ADDRESS_QUEUE:
            for i in range(count, settings.CC_ADDRESS_QUEUE):
                try:
                    Address.objects.create(address=coin.getnewaddress(settings.CC_ACCOUNT), currency=currency)
                except (socket_error, CannotSendRequest) :
                    pass


@shared_task()
def process_withdraw_transactions(ticker=None):
    if not ticker:
        for c in Currency.objects.all():
            process_withdraw_transactions.delay(c.ticker)
        return

    with transaction.atomic():
        currency = Currency.objects.select_for_update().get(ticker=ticker)
        coin = AuthServiceProxy(currency.api_url)

        # this will fail if bitcoin offline
        coin.getbalance()

        wtxs = WithdrawTransaction.objects.select_for_update() \
            .select_related('wallet') \
            .filter(currency=currency, state=WithdrawTransaction.NEW, txid=None) \
            .order_by('wallet')

        transaction_hash = {}
        for tx in wtxs:
            if tx.address in transaction_hash:
                transaction_hash[tx.address] += tx.amount
            else:
                transaction_hash[tx.address] = tx.amount

        if currency.dust > Decimal('0'):
            for address, amount in list(transaction_hash.items()):
                if amount < currency.dust:
                    wtxs = wtxs.exclude(currency=currency, address=address)
                    del transaction_hash[address]

        if not transaction_hash:
            return

        wtxs_ids = list(wtxs.values_list('id', flat=True))
        wtxs.update(state=WithdrawTransaction.ERROR)

    # this will fail if bitcoin offline
    coin.getbalance()

    txid = coin.sendmany(settings.CC_ACCOUNT, transaction_hash)

    if not txid:
        raise AssertionError('txid is empty')

    fee = coin.gettransaction(txid).get('fee', 0) * -1

    with transaction.atomic():
        currency = Currency.objects.select_for_update().get(ticker=ticker)
        wtxs = WithdrawTransaction.objects.select_for_update().filter(id__in=wtxs_ids)
        if not fee:
            fee_per_tx = 0
        else:
            fee_per_tx = (fee / len(wtxs))

        fee_hash = defaultdict(lambda : {'fee': Decimal('0'), 'amount': Decimal('0')})

        for tx in wtxs:
            fee_hash[tx.wallet]['fee'] += fee_per_tx
            fee_hash[tx.wallet]['amount'] += tx.amount

        for (wallet, data) in fee_hash.items():
            Operation.objects.create(
                wallet=wallet,
                holded=-data['amount'],
                balance=-data['fee'],
                description='Network fee',
                reason=tx
            )

            wallet = Wallet.objects.get(id=tx.wallet.id)
            wallet.balance -= data['fee']
            wallet.holded -= data['amount']
            wallet.save()

        wtxs.update(txid=txid, fee=fee_per_tx, state=WithdrawTransaction.DONE)



import string
import random
from decimal import Decimal
from mock import patch, MagicMock

from django.test import TransactionTestCase

from cc.models import Wallet, Address, Currency, Operation, Transaction, WithdrawTransaction
from cc import tasks
from cc import settings


settings.CC_CONFIRMATIONS = 2
settings.CC_ACCOUNT = ''


class DepositeTransaction(TransactionTestCase):
    def setUp(self):
        self.txdict = {
            'category': 'receive',
            'account': '',
            'blockhash': '000000008468ce0ec51c64aa98cf99b317274c5287e770c9eddaa83fa33222ef',
            'blockindex': 1,
            'walletconflicts': [],
            'time': 1409150845,
            'txid': '63fadb05b2f6b0c83925d402c6cf27bc841acaa8c89a335914f77f75b22ef5dc',
            'blocktime': 1409154118,
            'amount': Decimal('5.00000000'),
            'confirmations': 87,
            'timereceived': 1409150845,
            'address': 'mmxv3wYKozehzp3GZSUiKvRCWSJecWNSrd'
        }
        self.currency = Currency.objects.create(label='Bitcoin', ticker='btc')
        self.wallet = Wallet.objects.create(currency=self.currency, label='Test')
        self.address = Address.objects.create(address=self.txdict['address'], wallet=self.wallet, currency=self.currency)
        tasks.process_deposite_transaction(self.txdict, 'btc')

    def test_balance(self):
        wallet = Wallet.objects.get(id=self.wallet.id)
        self.assertEqual(wallet.balance, self.txdict['amount'])

    def test_tx(self):
        tx = Transaction.objects.get(txid=self.txdict['txid'])
        self.assertTrue(tx.processed)


class UnconfirmedTransaction(TransactionTestCase):
    def setUp(self):
        self.txdict = {
            'category': 'receive',
            'account': '',
            'blockhash': '000000008468ce0ec51c64aa98cf99b317274c5287e770c9eddaa83fa33222ef',
            'blockindex': 1,
            'walletconflicts': [],
            'time': 1409150845,
            'txid': '63fadb05b2f6b0c83925d402c6cf27bc841acaa8c89a335914f77f75b22ef5dc',
            'blocktime': 1409154118,
            'amount': Decimal('5.00000000'),
            'confirmations': 1,
            'timereceived': 1409150845,
            'address': 'mmxv3wYKozehzp3GZSUiKvRCWSJecWNSrd'
        }
        self.currency = Currency.objects.create(label='Bitcoin', ticker='btc')
        self.wallet = Wallet.objects.create(currency=self.currency, label='Test')
        self.address = Address.objects.create(address=self.txdict['address'], wallet=self.wallet, currency=self.currency)
        tasks.process_deposite_transaction(self.txdict, 'btc')

    def test_balance(self):
        wallet = Wallet.objects.get(id=self.wallet.id)
        self.assertEqual(wallet.unconfirmed, self.txdict['amount'])
        self.assertEqual(wallet.balance, Decimal('0'))

    def test_tx(self):
        tx = Transaction.objects.get(txid=self.txdict['txid'])
        self.assertFalse(tx.processed)


class ImmatureTransaction(TransactionTestCase):
    def setUp(self):
        self.txdict = {
            'category': 'immature',
            'account': '',
            'blockhash': '000000008468ce0ec51c64aa98cf99b317274c5287e770c9eddaa83fa33222ef',
            'blockindex': 1,
            'walletconflicts': [],
            'time': 1422794361,
            'txid': '63fadb05b2f6b0c83925d402c6cf27bc841acaa8c89a335914f77f75b22ef5dc',
            'blocktime': 1409154118,
            'amount': Decimal('1.00000000'),
            'confirmations': 1,
            'timereceived': 1422794579,
            'address': '12aJBBXWcWPxfQbXL4PQ3qtx978wwLL2g9'
        }
        self.currency = Currency.objects.create(label='Bitcoin', ticker='btc')
        self.wallet = Wallet.objects.create(currency=self.currency, label='Test')
        self.address = Address.objects.create(address=self.txdict['address'], wallet=self.wallet, currency=self.currency)
        tasks.process_deposite_transaction(self.txdict, 'btc')

    def test_balance(self):
        wallet = Wallet.objects.get(id=self.wallet.id)
        self.assertEqual(wallet.unconfirmed, self.txdict['amount'])
        self.assertEqual(wallet.balance, Decimal('0'))

    def test_tx(self):
        tx = Transaction.objects.get(txid=self.txdict['txid'])
        self.assertFalse(tx.processed)


class ConfirmTransaction(TransactionTestCase):
    def setUp(self):
        self.txdict = {
            'category': 'receive',
            'account': '',
            'blockhash': '000000008468ce0ec51c64aa98cf99b317274c5287e770c9eddaa83fa33222ef',
            'blockindex': 1,
            'walletconflicts': [],
            'time': 1409150845,
            'txid': '63fadb05b2f6b0c83925d402c6cf27bc841acaa8c89a335914f77f75b22ef5dc',
            'blocktime': 1409154118,
            'amount': Decimal('5.00000000'),
            'confirmations': 3,
            'timereceived': 1409150845,
            'address': 'mmxv3wYKozehzp3GZSUiKvRCWSJecWNSrd'
        }
        self.currency = Currency.objects.create(label='Bitcoin', ticker='btc')
        self.wallet = Wallet.objects.create(currency=self.currency, label='Test', unconfirmed=Decimal('5.0'))
        self.address = Address.objects.create(address=self.txdict['address'], wallet=self.wallet, currency=self.currency)
        self.tx = Transaction.objects.create(txid=self.txdict['txid'], address=self.txdict['address'], currency=self.currency)
        tasks.process_deposite_transaction(self.txdict, 'btc')

    def test_balance(self):
        wallet = Wallet.objects.get(id=self.wallet.id)
        self.assertEqual(wallet.balance, Decimal('5.0'))
        self.assertEqual(wallet.unconfirmed, Decimal('0'))

    def test_tx(self):
        tx = Transaction.objects.get(id=self.tx.id)
        self.assertTrue(tx.processed)


class WalletAddress(TransactionTestCase):
    def setUp(self):
        self.btc = Currency.objects.create(label='Bitcoin', ticker='btc')
        self.ltc = Currency.objects.create(label='Litecoin', ticker='ltc', magicbyte='48')
        self.wallet = Wallet.objects.create(currency=self.btc, label='Test')

    def test_no_addresses(self):
        self.assertIsNone(self.wallet.get_address())

    def test_active_address(self):
        active = Address.objects.create(address='1111111111111111111114oLvT2', wallet=self.wallet, currency=self.btc, active=True)
        Address.objects.create(address='1AGNa15ZQXAZUgFiqJ2i7Z2DPU2J6hW62i', wallet=self.wallet, currency=self.btc, active=False)
        self.assertEqual(self.wallet.get_address(), active)

    def test_unused_address(self):
        unused = Address.objects.create(address='1Eym7pyJcaambv8FG4ZoU8A4xsiL9us2zz', wallet=self.wallet, currency=self.btc, active=False)
        Address.objects.create(address='LRNYxwQsHpm2A1VhawrJQti3nUkPN7vtq3', currency=self.ltc, active=True)
        self.assertEqual(self.wallet.get_address(), unused)


class WalletWithdraw(TransactionTestCase):
    def setUp(self):
        self.btc = Currency.objects.create(label='Bitcoin', ticker='btc')
        self.wallet = Wallet.objects.create(currency=self.btc, label='Test', balance=Decimal('1.0'))
        self.amount = Decimal('1.0')
        self.address = '1111111111111111111114oLvT2'
        self.desc = 'some desc'

    def test_operation(self):
        self.wallet.withdraw_to_address(self.address, self.amount, self.desc)
        op = Operation.objects.all()[0]

        self.assertEqual(op.wallet, self.wallet)
        self.assertEqual(op.holded, self.amount)
        self.assertEqual(op.balance, -self.amount)
        self.assertEqual(op.description, self.desc)
        self.assertIsNotNone(op.reason)

    def test_wallet(self):
        self.wallet.withdraw_to_address(self.address, self.amount, self.desc)
        wallet = Wallet.objects.get(id=self.wallet.id)
        self.assertEqual(wallet.balance, Decimal('0'))
        self.assertEqual(wallet.holded, self.amount)

    def test_no_money(self):
        with self.assertRaises(ValueError):
            self.wallet.withdraw_to_address(self.address, Decimal('100'))

    def test_wrong_address(self):
        with self.assertRaises(ValueError):
            self.wallet.withdraw_to_address('mz4ZbfKfU4SQWRDagkfX2TLAotpimAAVFE', Decimal('100'))


class TaskWithdraw(TransactionTestCase):
    def setUp(self):
        self.currency = Currency.objects.create(label='Testnet', ticker='tst', magicbyte='111,196')
        self.wallet = Wallet.objects.create(currency=self.currency, label='Test', balance=Decimal('1.0'))
        self.txid = 'ea12fb225a0665e6ca35ab3fd7a514c36d1d5028d99340931d745dab62c13f8a'

        self.mock = MagicMock(name='asp')
        self.mock.return_value = self.mock
        self.mock.sendmany.return_value = self.txid
        self.mock.gettransaction.return_value = {
            'fee': Decimal('-0.00010000'),
            'timereceived': 1410086093,
            'hex': '010000000181d61f31536c155f43149bfa1a1ed0cd45c504f82e27f8411f14e6b37f926c62000000006a47304402206ac1b06a2ab1ad05c7874728188c73d99de7b8185bda0bce28e06762ea265ec0022050e614fe7aa4d34b75e6465685f12e3589a34763afa383414aaf4d2d62171ba20121020e871b50e2e1d46cac2f2b7611f1fcc97d41e9ced660682191c6ea391c7189e5ffffffff04e0247c38000000001976a91462403a00c6e4906d7624818eae2dc7572f0f592588ac002d3101000000001976a914a17b7337c17ab511686649515f7861944035846588ac80969800000000001976a91437138adaa16d895739a61d10d33fd0b898db552288ac80969800000000001976a914a621b489961d24092eba3838d3142173a8f9d8c488ac00000000',
            'txid': 'ea12fb225a0665e6ca35ab3fd7a514c36d1d5028d99340931d745dab62c13f8a',
            'amount': Decimal('-0.40000000'),
            'walletconflicts': [],
            'details': [{'category': 'send',
            'account': '',
            'fee': Decimal('-0.00010000'),
            'amount': Decimal('-0.20000000'),
            'address': 'mvEnyQ9b9iTA11QMHAwSVtHUrtD4CTfiDB'},
            {'category': 'send',
            'account': '',
            'fee': Decimal('-0.00010000'),
            'amount': Decimal('-0.10000000'),
            'address': 'mkYAsS9QLYo5mXVjuvxKkZUhQJxiMLX5Xk'},
            {'category': 'send',
            'account': '',
            'fee': Decimal('-0.00010000'),
            'amount': Decimal('-0.10000000'),
            'address': 'mvfNqn5AoVWrsJGuKrdPuoQhYs71CR9uFA'}],
            'confirmations': 0,
            'time': 1410086093
        }


    def test_process_withdraw_transactions(self):
        self.wallet.withdraw_to_address('mvEnyQ9b9iTA11QMHAwSVtHUrtD4CTfiDB', Decimal('0.1'))
        self.wallet.withdraw_to_address('mkYAsS9QLYo5mXVjuvxKkZUhQJxiMLX5Xk', Decimal('0.1'))
        self.wallet.withdraw_to_address('mvEnyQ9b9iTA11QMHAwSVtHUrtD4CTfiDB', Decimal('0.1'))
        self.wallet.withdraw_to_address('mvfNqn5AoVWrsJGuKrdPuoQhYs71CR9uFA', Decimal('0.1'))

        with patch('cc.tasks.AuthServiceProxy', self.mock):
            tasks.process_withdraw_transactions(ticker=self.currency.ticker)

        self.mock.gettransaction.assert_called_once_with(self.txid)
        self.mock.sendmany.assert_called_once_with('', {
            'mvEnyQ9b9iTA11QMHAwSVtHUrtD4CTfiDB': Decimal('0.2'),
            'mkYAsS9QLYo5mXVjuvxKkZUhQJxiMLX5Xk': Decimal('0.1'),
            'mvfNqn5AoVWrsJGuKrdPuoQhYs71CR9uFA': Decimal('0.1')
        })

        wallet = Wallet.objects.get(id=self.wallet.id)
        self.assertEqual(wallet.holded, Decimal('0'))
        self.assertEqual(wallet.balance, Decimal('0.5999'))

        fee_operation = Operation.objects.get(wallet=wallet, description='Network fee')
        self.assertEqual(fee_operation.balance, Decimal('-0.0001'))
        self.assertEqual(fee_operation.holded, Decimal('-0.4'))


class TaskRefillAddressQueue(TransactionTestCase):
    def setUp(self):
        self.currency = Currency.objects.create(label='Testnet', ticker='tst', magicbyte='111,196')
        self.wallet = Wallet.objects.create(currency=self.currency)

        self.mock = MagicMock(name='asp')
        self.mock.return_value = self.mock
        self.mock.getnewaddress.side_effect = lambda: ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(20))

    def refill_addresses_queue(self):
        self.assertEqual(len(Address.objects.all()), 0)

        with patch('cc.tasks.AuthServiceProxy', self.mock):
            tasks.refill_addresses_queue()

        self.assertEqual(len(Address.objects.all()), settings.CC_ADDRESS_QUEUE)


class WalletTransfer(TransactionTestCase):
    def setUp(self):
        self.currency = Currency.objects.create(label='Testnet', ticker='tst', magicbyte='111,196')
        self.wallet1 = Wallet.objects.create(currency=self.currency, balance=1)
        self.wallet2 = Wallet.objects.create(currency=self.currency, balance=0)

    def test_transfer(self):
        self.wallet1.transfer(Decimal('1'), self.wallet2)

        self.assertEqual(self.wallet1.balance, 0)
        self.assertEqual(self.wallet2.balance, 1)

        o1 = Operation.objects.get(wallet=self.wallet1)
        o2 = Operation.objects.get(wallet=self.wallet2)

        self.assertEqual(o1.balance, -1)
        self.assertEqual(o2.balance, 1)


class QueryTransaction(TransactionTestCase):
    def setUp(self):
        self.txdict = {
            "amount" : Decimal('3'),
            "confirmations" : 54271,
            "blockhash" : "00000000000000017adae0a02c36e7a6379991aad2ce35c2ba1b540aff01d7b8",
            "blockindex" : 744,
            "blocktime" : 1391222639,
            "txid" : "01c17411ff6a4278ada87c28dad74b9d1e79c799743fd2d63dac945645123ab3",
            "walletconflicts" : [
            ],
            "time" : 1391222405,
            "timereceived" : 1391222405,
            "details" : [
                {
                    "account" : "somerandomstring14aqqwd",
                    "address" : "16ahqjUA7VJMuBpKjR3zX48xnTgPMM47cr",
                    "category" : "receive",
                    "amount" : Decimal('1')
                },
                {
                    "account" : "somerandomstring14aqqwd",
                    "address" : "1FLrCWUJw5SG7uDHzkrRLih55PxMC763eu",
                    "category" : "receive",
                    "amount" : Decimal('2')
                }
            ]
        }
        self.currency = Currency.objects.create(label='BItcoin', ticker='BTC', magicbyte='0,5')
        self.wallet = Wallet.objects.create(currency=self.currency, label='Test')
        self.address1 = Address.objects.create(address=self.txdict['details'][0]['address'], wallet=self.wallet, currency=self.currency)
        self.address2 = Address.objects.create(address=self.txdict['details'][1]['address'], wallet=self.wallet, currency=self.currency)

        self.mock = MagicMock(name='asp')
        self.mock.return_value = self.mock
        self.mock.gettransaction.return_value = self.txdict

    def test_balance(self):
        with patch('cc.tasks.AuthServiceProxy', self.mock):
            tasks.query_transaction('BTC', self.txdict['txid'])

        wallet = Wallet.objects.get(id=self.wallet.id)

        self.assertEqual(wallet.balance, Decimal('3'))


class Dust(TransactionTestCase):
    def setUp(self):
        self.currency = Currency.objects.create(label='Testnet', ticker='tst', magicbyte='111,196', dust=Decimal('0.00005430'))
        self.wallet = Wallet.objects.create(currency=self.currency, label='Test', balance=Decimal('2.0'))
        self.txid = 'ea12fb225a0665e6ca35ab3fd7a514c36d1d5028d99340931d745dab62c13f8a'

        self.mock = MagicMock(name='asp')
        self.mock.return_value = self.mock
        self.mock.sendmany.return_value = self.txid
        self.mock.gettransaction.return_value = {
            'fee': Decimal('-0.00010000'),
            'timereceived': 1410086093,
            'hex': '010000000181d61f31536c155f43149bfa1a1ed0cd45c504f82e27f8411f14e6b37f926c62000000006a47304402206ac1b06a2ab1ad05c7874728188c73d99de7b8185bda0bce28e06762ea265ec0022050e614fe7aa4d34b75e6465685f12e3589a34763afa383414aaf4d2d62171ba20121020e871b50e2e1d46cac2f2b7611f1fcc97d41e9ced660682191c6ea391c7189e5ffffffff04e0247c38000000001976a91462403a00c6e4906d7624818eae2dc7572f0f592588ac002d3101000000001976a914a17b7337c17ab511686649515f7861944035846588ac80969800000000001976a91437138adaa16d895739a61d10d33fd0b898db552288ac80969800000000001976a914a621b489961d24092eba3838d3142173a8f9d8c488ac00000000',
            'txid': 'ea12fb225a0665e6ca35ab3fd7a514c36d1d5028d99340931d745dab62c13f8a',
            'amount': Decimal('-10000000'),
            'walletconflicts': [],
            'details': [{'category': 'send',
            'account': '',
            'fee': Decimal('-0.00010000'),
            'amount': Decimal('-1.0000000'),
            'address': 'mvEnyQ9b9iTA11QMHAwSVtHUrtD4CTfiDB'}]
        }

    def test_process_withdraw_transactions(self):
        self.wallet.withdraw_to_address('mvEnyQ9b9iTA11QMHAwSVtHUrtD4CTfiDB', Decimal('1'))
        self.wallet.withdraw_to_address('mvfNqn5AoVWrsJGuKrdPuoQhYs71CR9uFA', Decimal('0.00000001'))

        with patch('cc.tasks.AuthServiceProxy', self.mock):
            tasks.process_withdraw_transactions(ticker=self.currency.ticker)

        self.mock.sendmany.assert_called_once_with('', {
            'mvEnyQ9b9iTA11QMHAwSVtHUrtD4CTfiDB': Decimal('1')
        })

        wallet = Wallet.objects.get(id=self.wallet.id)
        self.assertEqual(wallet.balance, Decimal('0.99989999'))

        wt1 = WithdrawTransaction.objects.get(address='mvEnyQ9b9iTA11QMHAwSVtHUrtD4CTfiDB')
        wt2 = WithdrawTransaction.objects.get(address='mvfNqn5AoVWrsJGuKrdPuoQhYs71CR9uFA')
        self.assertEqual(wt1.txid, self.txid)
        self.assertIsNone(wt2.txid)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-05-22 08:08
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('cc', '0008_withdrawtransaction_state'),
    ]

    operations = [
        migrations.AlterField(
            model_name='withdrawtransaction',
            name='txid',
            field=models.CharField(blank=True, db_index=True, max_length=100, null=True, verbose_name='Txid'),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('cc', '0003_auto_20150106_1041'),
    ]

    operations = [
        migrations.AddField(
            model_name='transaction',
            name='address',
            field=models.CharField(default='', max_length=50, verbose_name='Address'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='transaction',
            name='txid',
            field=models.CharField(max_length=100, verbose_name='Txid'),
        ),
        migrations.AlterUniqueTogether(
            name='transaction',
            unique_together=set([('txid', 'address')]),
        ),
    ]



from decimal import Decimal as D

from cc.models import Wallet, Operation, Address, Currency, Transaction, WithdrawTransaction


def total_recieved(ticker, listreceivedbyaddress):
    currency = Currency.objects.get(ticker=ticker)
    wallets = Wallet.objects.filter(currency=currency)

    data = dict(map(lambda x: [x['address'], x['amount']], listreceivedbyaddress))

    result = {'mismatch': [], 'missing': []}

    for w in wallets:
        tr = w.total_received()
        if tr > 0:
            summ = D('0')
            for a in Address.objects.filter(wallet = w):
                try:
                    summ += data[a.address]
                except KeyError:
                    result['missing'].append({'address': a.address})

            if (tr != summ):
                result['mismatch'].append({
                    'wallet': w.id,
                    'db': tr,
                    'coin': summ,
                })

    return result


def double_spend(ticker, listtransactions, ccc):
    currency = Currency.objects.get(ticker=ticker)


    send = filter(lambda x: x['category'] == 'send', listtransactions)
    missing = []

    for tx in send:
        wtxs = WithdrawTransaction.objects.filter(txid=tx['txid'], currency=currency)

        if not wtxs:
            ccc.stdout.write(ccc.style.ERROR('Missing %(txid)s  %(address)s  %(amount)s  %(time)s' % tx))
            missing.append(tx)



    import IPython; IPython.embed()
    return

    return result



from django.conf.urls import url

import cc.views as views


urlpatterns = [
    url(r'^blocknotify/$', views.blocknotify, name='cc-blocknotify'),
    url(r'^walletnotify/$', views.walletnotify, name='cc-walletnotify'),
]



from __future__ import absolute_import

from django import forms

from .models import Wallet

class WalletAdminForm(forms.ModelForm):
    class Meta:
        model = Wallet
        exclude = []

    def clean_label(self):
        return self.cleaned_data['label'] or None



# -*- coding: utf-8 -*-
from decimal import Decimal

from django.db import models
from django.db.models import Sum

from django.utils.translation import ugettext_lazy as _
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.utils.timezone import now

from cc import settings
from cc.validator import validate

class Wallet(models.Model):
    currency = models.ForeignKey('Currency')
    balance = models.DecimalField(_('Balance'), max_digits=18, decimal_places=8, default=0)
    holded = models.DecimalField(_('Holded'), max_digits=18, decimal_places=8, default=0)
    unconfirmed = models.DecimalField(_('Unconfirmed'), max_digits=18, decimal_places=8, default=0)
    label = models.CharField(_('Label'), max_length=100, blank=True, null=True, unique=True)

    def __str__(self):
        return u'{0} {1} "{2}"'.format(self.balance, self.currency.ticker, self.label or '')

    def get_address(self):
        active = Address.objects.filter(wallet=self, active=True, currency=self.currency)[:1]
        if active:
            return active[0]

        unused = Address.objects.filter(wallet=None, active=True, currency=self.currency)[:1]
        if unused:
            free = unused[0]
            free.wallet = self
            free.save()
            return free

        old = Address.objects.filter(wallet=self, active=False, currency=self.currency)[:1]
        if old:
            return old[0]

    def withdraw(self, amount, description="", reason=None):
        if amount < 0:
            raise ValueError('Invalid amount')

        if self.balance - amount < -settings.CC_ALLOW_NEGATIVE_BALANCE:
            raise ValueError('No money')

        Operation.objects.create(
            wallet=self,
            balance=-amount,
            description=description,
            reason=reason
        )
        self.balance -= amount
        self.save()

    def transfer(self, amount, deposite_wallet, reason=None, description=""):
        if amount < 0:
            raise ValueError('Invalid amount')

        if self.balance - amount < -settings.CC_ALLOW_NEGATIVE_BALANCE:
            raise ValueError('No money')

        Operation.objects.create(
            wallet=self,
            balance=-amount,
            description=description,
            reason=reason
        )
        Operation.objects.create(
            wallet=deposite_wallet,
            balance=+amount,
            description=description,
            reason=reason
        )
        self.balance -= amount
        self.save()
        deposite_wallet.balance += amount
        deposite_wallet.save()

    def withdraw_to_address(self, address, amount, description=""):
        if not validate(address, self.currency.magicbyte):
            raise ValueError('Invalid address')

        if amount < 0:
            raise ValueError('Invalid amount')

        if self.balance - amount < -settings.CC_ALLOW_NEGATIVE_BALANCE:
            raise ValueError('No money')

        tx = WithdrawTransaction.objects.create(
            currency=self.currency,
            amount=amount,
            address=address,
            wallet=self,
        )
        op = Operation.objects.create(
            wallet=self,
            balance=-amount,
            holded=amount,
            description=description,
            reason=tx
        )
        self.balance -= amount
        self.holded += amount
        self.save()

        return {
            'tx': tx,
            'op': op,
        }

    def total_received(self):
        return Operation.objects.filter(wallet=self, balance__gt=0).aggregate(balance=Sum('balance'))['balance'] or Decimal('0')

    def recalc_balance(self, save=False):
        recalc = Operation.objects.filter(wallet=self).aggregate(balance=Sum('balance'),
                                   holded=Sum('holded'),
                                   unconfirmed=Sum('unconfirmed'))

        for k, v in recalc.items():
            if v is None:
                recalc[k] = Decimal('0')

        if save:
            self.balance = recalc['balance']
            self.holded = recalc['holded']
            self.unconfirmed = recalc['unconfirmed']
            self.save()

        return recalc

    def get_operations(self):
        return Operation.objects.filter(wallet=self).order_by('-created')

    def get_unpaid_dust_summary(self):
        if not self.currency.dust:
            return {}

        txs = WithdrawTransaction.objects.filter(wallet=self, txid=None, amount__lt=self.currency.dust)
        if len(txs) == 0:
            return {}

        from collections import defaultdict
        tx_hash = defaultdict(lambda : Decimal('0'))
        for tx in txs:
            tx_hash[tx.address] += tx.amount

        return dict(tx_hash)


class Operation(models.Model):
    wallet = models.ForeignKey(Wallet)
    created = models.DateTimeField(_('Created'), default=now)
    balance = models.DecimalField(_('Balance'), max_digits=18, decimal_places=8, default=0)
    holded = models.DecimalField(_('Holded'), max_digits=18, decimal_places=8, default=0)
    unconfirmed = models.DecimalField(_('Unconfirmed'), max_digits=18, decimal_places=8, default=0)

    description = models.CharField(_('Description'), max_length=100, blank=True, null=True)

    reason_content_type = models.ForeignKey(ContentType, null=True, blank=True)
    reason_object_id = models.PositiveIntegerField(null=True, blank=True)
    reason = GenericForeignKey('reason_content_type', 'reason_object_id')


class Address(models.Model):
    address = models.CharField(_('Address'), max_length=50, primary_key=True)
    currency = models.ForeignKey('Currency')
    created = models.DateTimeField(_('Created'), default=now)
    active = models.BooleanField(_('Active'), default=True)
    label = models.CharField(_('Label'), max_length=50, blank=True, null=True, default=None)
    wallet = models.ForeignKey(Wallet, blank=True, null=True, related_name="addresses")

    def __str__(self):
        return u'{0}, {1}'.format(self.address, self.currency.ticker)


class Currency(models.Model):
    ticker = models.CharField(_('Ticker'), max_length=4, default='BTC', primary_key=True)
    label = models.CharField(_('Label'), max_length=20, default='Bitcoin', unique=True)
    magicbyte = models.CommaSeparatedIntegerField(_('Magicbytes'), max_length=10, default='0,5')
    last_block = models.PositiveIntegerField(_('Last block'), blank=True, null=True, default=0)
    api_url = models.CharField(_('API hostname'), default='http://localhost:8332', max_length=100, blank=True, null=True)
    dust = models.DecimalField(_('Dust'), max_digits=18, decimal_places=8, default=Decimal('0.0000543'))

    class Meta:
        verbose_name_plural = _('currencies')

    def __str__(self):
        return self.label


class Transaction(models.Model):
    txid = models.CharField(_('Txid'), max_length=100)
    address = models.CharField(_('Address'), max_length=50)
    currency = models.ForeignKey('Currency')
    processed = models.BooleanField(_('Processed'), default=False)

    class Meta:
        unique_together = (('txid', 'address'),)


class WithdrawTransaction(models.Model):
    NEW = 'NEW'
    ERROR = 'ERROR'
    DONE = 'DONE'
    WTX_STATES = (
        ('NEW', 'New'),
        ('ERROR', 'Error'),
        ('DONE', 'Done'),
    )
    currency = models.ForeignKey('Currency')
    amount = models.DecimalField(_('Amount'), max_digits=18, decimal_places=8)
    address = models.CharField(_('Address'), max_length=50)
    wallet = models.ForeignKey(Wallet)
    created = models.DateTimeField(_('Created'), default=now)
    txid = models.CharField(_('Txid'), max_length=100, blank=True, null=True, db_index=True)
    walletconflicts = models.CharField(_('Walletconflicts txid'), max_length=100, blank=True, null=True, db_index=True)
    state = models.CharField(_('State'), max_length=10, choices=WTX_STATES, default=NEW)
    fee = models.DecimalField(_('Fee'), max_digits=18, decimal_places=8, null=True, blank=True)



from __future__ import absolute_import

from django.http import HttpResponse, HttpResponseForbidden, HttpResponseBadRequest
from django.http.request import validate_host, split_domain_port

from . import settings
from .models import Currency
from .tasks import query_transactions, query_transaction


def cc_validate_host(func):
    def validate(request):
        domain, port = split_domain_port(request.META['HTTP_HOST'])
        if not validate_host(domain, settings.CC_ALLOWED_HOSTS):
            return HttpResponseForbidden('forbiden')

        return func(request)

    return validate


def vaidate_currency(func):
    def validate(request):
        ticker = request.GET.get('currency')
        if not ticker:
            return HttpResponseBadRequest('Currency is missing')

        try:
            currency = Currency.objects.get(ticker=ticker)
        except Currency.DoesNotExist:
            return HttpResponseBadRequest('Wrong currency ticker')

        return func(request)

    return validate


def vaidate_txid(func):
    def validate(request):
        if not request.GET.get('txid'):
            return HttpResponseBadRequest('Txid is missing')

        return func(request)

    return validate


@cc_validate_host
@vaidate_currency
def blocknotify(request):
    query_transactions.delay(ticker=request.GET['currency'])
    return HttpResponse('success')


@cc_validate_host
@vaidate_currency
@vaidate_txid
def walletnotify(request):
    query_transaction.delay(request.GET['currency'], request.GET['txid'])
    return HttpResponse('success')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('cc', '0002_auto_20150105_1859'),
    ]

    operations = [
        migrations.AlterField(
            model_name='withdrawtransaction',
            name='fee',
            field=models.DecimalField(null=True, verbose_name='Fee', max_digits=18, decimal_places=8, blank=True),
        ),
    ]



from setuptools import setup, find_packages
from os.path import join, dirname


setup(
    author='Ivan Vershigora',
    author_email='ivan.vershigora@gmail.com',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Build Tools',
    ],
    description='Django wallet for Bitcoin and other cryptocurrencies',
    download_url = 'https://github.com/limpbrains/django-cc/tarball/0.2.1',
    install_requires=[
        'celery>=3,<4',
        'Django>=1.7',
        'mock',
        'pycoin',
        'python-bitcoinrpc>=1.0',
    ],
    keywords='bitcoin litecoin django wallet cryptocurrency',
    license='MIT License',
    long_description=open(join(dirname(__file__), 'README.md')).read(),
    name='django-cc',
    packages=find_packages(),
    url='https://github.com/limpbrains/django-cc',
    version='0.2.1',
)



from django.contrib import admin

from . import models
from .forms import WalletAdminForm


class WalletAdmin(admin.ModelAdmin):
    form = WalletAdminForm
    list_display = ('id', 'currency', 'balance', 'holded', 'unconfirmed', 'label', 'get_address')
    list_filter = ('currency',)

admin.site.register(models.Wallet, WalletAdmin)


class OperationAdmin(admin.ModelAdmin):
    list_display = ('id', 'wallet', 'balance', 'holded', 'unconfirmed', 'description')
    list_filter = ('wallet',)

admin.site.register(models.Operation, OperationAdmin)


class CurrencyAdmin(admin.ModelAdmin):
    list_display = ('ticker', 'label', 'last_block')

admin.site.register(models.Currency, CurrencyAdmin)


class TransactionAdmin(admin.ModelAdmin):
    list_display = ('id', 'txid', 'currency', 'processed')
    list_filter = ('currency', 'processed')

admin.site.register(models.Transaction, TransactionAdmin)


class AddressAdmin(admin.ModelAdmin):
    list_display = ('address', 'currency', 'created', 'active', 'label', 'wallet')
    list_filter = ('currency', 'active')

admin.site.register(models.Address, AddressAdmin)


class WithdrawTransactionAdmin(admin.ModelAdmin):
    list_display = ('id', 'currency', 'amount', 'address', 'wallet', 'created', 'state', 'txid', 'walletconflicts', 'fee')
    list_filter = ('currency', 'state')

admin.site.register(models.WithdrawTransaction, WithdrawTransactionAdmin)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Address',
            fields=[
                ('address', models.CharField(max_length=50, serialize=False, verbose_name='Address', primary_key=True)),
                ('created', models.DateTimeField(default=django.utils.timezone.now, verbose_name='Created')),
                ('active', models.BooleanField(default=True, verbose_name='Active')),
                ('label', models.CharField(default=None, max_length=50, null=True, verbose_name='Label', blank=True)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Currency',
            fields=[
                ('ticker', models.CharField(default=b'BTC', max_length=4, serialize=False, verbose_name='Ticker', primary_key=True)),
                ('label', models.CharField(default=b'Bitcoin', unique=True, max_length=20, verbose_name='Label')),
                ('magicbyte', models.CommaSeparatedIntegerField(default=b'0,5', max_length=10, verbose_name='Magicbytes')),
                ('last_block', models.PositiveIntegerField(default=0, null=True, verbose_name='Last block', blank=True)),
                ('api_url', models.CharField(default=b'http://localhost:8332', max_length=100, null=True, verbose_name='API hostname', blank=True)),
            ],
            options={
                'verbose_name_plural': 'currencies',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Operation',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created', models.DateTimeField(default=django.utils.timezone.now, verbose_name='Created')),
                ('balance', models.DecimalField(default=0, verbose_name='Balance', max_digits=18, decimal_places=8)),
                ('holded', models.DecimalField(default=0, verbose_name='Holded', max_digits=18, decimal_places=8)),
                ('unconfirmed', models.DecimalField(default=0, verbose_name='Unconfirmed', max_digits=18, decimal_places=8)),
                ('description', models.CharField(max_length=100, null=True, verbose_name='Description', blank=True)),
                ('reason_object_id', models.PositiveIntegerField(null=True, blank=True)),
                ('reason_content_type', models.ForeignKey(blank=True, to='contenttypes.ContentType', null=True)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Transaction',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('txid', models.CharField(unique=True, max_length=100, verbose_name='Txid')),
                ('processed', models.BooleanField(default=False, verbose_name='Processed')),
                ('currency', models.ForeignKey(to='cc.Currency')),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Wallet',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('balance', models.DecimalField(default=0, verbose_name='Balance', max_digits=18, decimal_places=8)),
                ('holded', models.DecimalField(default=0, verbose_name='Holded', max_digits=18, decimal_places=8)),
                ('unconfirmed', models.DecimalField(default=0, verbose_name='Unconfirmed', max_digits=18, decimal_places=8)),
                ('label', models.CharField(max_length=100, null=True, verbose_name='Label', blank=True)),
                ('currency', models.ForeignKey(to='cc.Currency')),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='WithdrawTransaction',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('amount', models.DecimalField(verbose_name='Amount', max_digits=18, decimal_places=8)),
                ('address', models.CharField(max_length=50, verbose_name='Address')),
                ('created', models.DateTimeField(default=django.utils.timezone.now, verbose_name='Created')),
                ('txid', models.CharField(max_length=100, null=True, verbose_name='Txid', blank=True)),
                ('fee', models.DecimalField(null=True, verbose_name='Amount', max_digits=18, decimal_places=8, blank=True)),
                ('currency', models.ForeignKey(to='cc.Currency')),
                ('wallet', models.ForeignKey(to='cc.Wallet')),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.AddField(
            model_name='operation',
            name='wallet',
            field=models.ForeignKey(to='cc.Wallet'),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='address',
            name='currency',
            field=models.ForeignKey(to='cc.Currency'),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='address',
            name='wallet',
            field=models.ForeignKey(related_name=b'addresses', blank=True, to='cc.Wallet', null=True),
            preserve_default=True,
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-05-22 08:05
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('cc', '0007_withdrawtransaction_state'),
    ]

    operations = [
        migrations.AlterField(
            model_name='withdrawtransaction',
            name='state',
            field=models.CharField(choices=[('NEW', 'New'), ('ERROR', 'Error'), ('DONE', 'Done')], default='NEW', max_length=10, verbose_name='State'),
        ),
    ]
